/*
 * Copyright (c) 2011, Andreas Spall
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Andreas Spall nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.osoboo.jira.metadata.internal;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import com.atlassian.activeobjects.external.ActiveObjects;
import com.atlassian.jira.project.ProjectImpl;
import com.osoboo.jira.metadata.JiraMetadata;
import com.osoboo.jira.metadata.Metadata;
import com.osoboo.jira.metadata.MetadataService;

public class MetadataServiceImpl implements MetadataService {

	private final ActiveObjects ao;
	@SuppressWarnings("rawtypes")
	private Map<Class, EnrichedObjectKeyGenerator> registeredGenerators = new HashMap<Class, EnrichedObjectKeyGenerator>();

	public MetadataServiceImpl(ActiveObjects ao) {
		this.ao = ao;
		registeredGenerators.put(ProjectImpl.class, new KeyGeneratorProject());
	}

	/** {@inheritDoc} */
	@Override
	public void save(Object enrichedObject, String key, String value) {
		String enrichedObjectKey = getEnrichedObjectKey(enrichedObject);
		JiraMetadata metadata = null;
		JiraMetadata[] result = ao.find(JiraMetadata.class, "ENRICHED_OBJECT_KEY = ? AND USER_KEY = ?", enrichedObjectKey, key);
		if (result.length > 0) {
			metadata = result[0];
		} else {
			metadata = ao.create(JiraMetadata.class);
			metadata.setEnrichedObjectKey(enrichedObjectKey);
			metadata.setUserKey(key);
		}
		metadata.setUserValue(value);
		metadata.save();
	}

	/** {@inheritDoc} */
	@Override
	public String getMetadataValue(Object enrichedObject, String key) {
		JiraMetadata metadata = getMetadata(enrichedObject, key);
		if (metadata == null) {
			return "";
		}
		return metadata.getUserValue();
	}

	/** {@inheritDoc} */
	@Override
	public Set<JiraMetadata> getMetadata(Object enrichedObject) {
		String enrichedObjectKey = getEnrichedObjectKey(enrichedObject);
		try {
			JiraMetadata[] result = ao.find(JiraMetadata.class, "ENRICHED_OBJECT_KEY = ?", enrichedObjectKey);
			if (result.length == 0) {
				cleanupLatencyData();
				result = ao.find(JiraMetadata.class, "ENRICHED_OBJECT_KEY = ?", enrichedObjectKey);
			}
			Set<JiraMetadata> resultSet = new HashSet<JiraMetadata>();
			for (JiraMetadata metadata : result) {
				resultSet.add(metadata);
			}
			return resultSet;
		} catch (Exception e) {
			return new HashSet<JiraMetadata>();
		}
	}

	/** {@inheritDoc} */
	@Override
	public void delete(Object enrichedObject, int id) {
		String enrichedObjectKey = getEnrichedObjectKey(enrichedObject);
		JiraMetadata[] result = ao.find(JiraMetadata.class, "ENRICHED_OBJECT_KEY = ? AND ID = ?", enrichedObjectKey, id);
		ao.delete(result);
	}

	/**
	 * copy the pre 1.0.2 data to the new structure.
	 */
	private void cleanupLatencyData() {
		try {
			Metadata[] find = ao.find(Metadata.class);
			for (Metadata metadata : find) {
				JiraMetadata jMetadata = ao.create(JiraMetadata.class);
				jMetadata.setEnrichedObjectKey(metadata.getEnrichedObjectKey());
				jMetadata.setUserKey(metadata.getKey());
				jMetadata.setUserValue(metadata.getValue());
				jMetadata.save();
				ao.delete(metadata);
			}
		} catch (Exception e) {
			// ignore it
		}

	}

	/**
	 * @param enrichedObject
	 *            the enriced object
	 * @param key
	 *            the metadata key
	 * @return the string value
	 */
	private JiraMetadata getMetadata(Object enrichedObject, String key) {
		String enrichedObjectKey = getEnrichedObjectKey(enrichedObject);
		JiraMetadata[] result = ao.find(JiraMetadata.class, "ENRICHED_OBJECT_KEY = ? AND USER_KEY = ?", enrichedObjectKey, key);
		if (result.length == 0) {
			cleanupLatencyData();
			result = ao.find(JiraMetadata.class, "ENRICHED_OBJECT_KEY = ? AND USER_KEY = ?", enrichedObjectKey, key);
		}
		if (result.length > 0) {
			return result[0];
		}
		return null;
	}

	/**
	 * @param enrichedObject
	 * @return the generated object key will be returned
	 */
	private String getEnrichedObjectKey(Object enrichedObject) {
		if (enrichedObject != null && registeredGenerators.containsKey(enrichedObject.getClass())) {
			return registeredGenerators.get(enrichedObject.getClass()).getKey(enrichedObject);
		}
		return null;
	}

}
